/*
    QuickMSG
    Copyright (C) 2014  Jeroen Vreeken <jeroen@vreeken.net>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/
#define _GNU_SOURCE
#include "smtp.h"
#include "mailbox.h"
#include "message.h"
#include "gpg.h"
#include "quickmsg.h"
#include "settings.h"
#include "contact.h"
#include "quickmsg_db.h"
#include "post.h"
#include "attachment_file.h"

#include <stdio.h>
#include <string.h>
#include <time.h>

#define QUICKMSG_MAILBOX "INBOX"

int quickmsg_checkdir(struct mailbox *mailbox)
{
	int r;
	
	r = mailbox_chdir(mailbox, QUICKMSG_MAILBOX);
	if (!r)
		return 0;
	
	r = mailbox_mkdir(mailbox, QUICKMSG_MAILBOX);
	if (!r)
		return 0;
	
	return 0;
}

void quickmsg_check_me(struct quickmsg_db *db)
{
	struct contact *c;
	char *me = settings_address_get();
	
	c = quickmsg_db_contact_get_by_id(db, 1);
	if (!c) {
		c = contact_alloc();
		contact_address_set(c, me);
		contact_name_set(c, settings_name_get());
		contact_type_set(c, CONTACT_TYPE_PERSON);
		contact_keystat_set(c, CONTACT_KEYSTAT_VERIFIED);
		contact_time_lastact_set(c, time(NULL));

		quickmsg_db_contact_insert(db, c);
	} else {
		if (strcmp(contact_address_get(c), me)) {
			contact_address_set(c, me);
			contact_name_set(c, settings_name_get());
			contact_type_set(c, CONTACT_TYPE_PERSON);
			contact_keystat_set(c, CONTACT_KEYSTAT_VERIFIED);
			contact_time_lastact_set(c, time(NULL));
			quickmsg_db_contact_update(db, c);
		}
	}
	
	contact_free(c);
}

void quickmsg_list(struct quickmsg_db *db)
{
	struct contact_list *l;
	size_t i;
	
	l = quickmsg_db_contact_get_all(db);
	
	printf(contact_print_header());
	for (i = 0; i < l->size; i++) {
		printf(contact_print(l->contacts[i]));
	}
	
	contact_list_free(l);
}

char *quickmsg_attachment_save(struct attachment *a)
{
	char *aname = a->name;
	int i = 0;
	FILE *fa;
	char *filename = NULL;
	
	if (!aname)
		aname = "attachment";
	
	do {
		if (i == 0) {
			asprintf(&filename, "%s%s",
			    settings_attachment_dirname_get(), aname);
		} else {
			free(filename);
			asprintf(&filename, "%s%s.%d",
			    settings_attachment_dirname_get(), aname, i);
		}
		fa = fopen(filename, "r");
		if (fa) {
			fclose(fa);
			fa = NULL;
			i++;
		} else {
			fa = fopen(filename, "w");
			if (!fa)
				return NULL;
		}
	} while (!fa);
	
	fwrite(a->data, a->length, 1, fa);
	fflush(fa);
	fclose(fa);

	return filename;
}

#define QUICKMSG_ACTION "Action: "
#define QUICKMSG_ACTION_POST "post"
#define QUICKMSG_ACTION_GROUP "group"
#define QUICKMSG_NAME "Name: "
#define QUICKMSG_TIME "Time: "
#define QUICKMSG_MEMBERS "Members: "
#define QUICKMSG_GROUP "Group: "
#define QUICKMSG_OWNER "Owner: "


void quickmsg_handle(struct quickmsg_db *db, char *from, struct attachment *qm, struct attachment *a)
{
	char *line, *end;
	size_t len, ulen;
	char *value;
	bool post = false;
	bool group = false;
	char *name = NULL;
	bool havetime = false;
	time_t acttime;
	char *owner = NULL;
	char *members = NULL;
	int groupnr;
	bool havegroupnr = false;
	char *postdata = NULL;
	
	end = qm->data;
	end += qm->length;
	
	for (line = qm->data; line < end; line += len + 1) {
		for (len = 0; line + len < end && line[len] != '\n'; len++);
		ulen = len;
		if (line[len - 1] == '\r')
			ulen--;
		
		if (!strncmp(line, QUICKMSG_ACTION, strlen(QUICKMSG_ACTION))) {
			value = line + strlen(QUICKMSG_ACTION);
			if (!strncmp(value, QUICKMSG_ACTION_POST, strlen(QUICKMSG_ACTION_POST)))
				post = true;
			if (!strncmp(value, QUICKMSG_ACTION_GROUP, strlen(QUICKMSG_ACTION_GROUP)))
				group = true;
		}
		if (!strncmp(line, QUICKMSG_NAME, strlen(QUICKMSG_NAME))) {
			value = line + strlen(QUICKMSG_NAME);
			
			name = calloc(1, ulen - strlen(QUICKMSG_NAME) + 1);
			memcpy(name, value, ulen - strlen(QUICKMSG_NAME));
		}
		if (!strncmp(line, QUICKMSG_OWNER, strlen(QUICKMSG_OWNER))) {
			value = line + strlen(QUICKMSG_OWNER);
			
			owner = calloc(1, ulen - strlen(QUICKMSG_OWNER) + 1);
			memcpy(owner, value, ulen - strlen(QUICKMSG_OWNER));
		}
		if (!strncmp(line, QUICKMSG_MEMBERS, strlen(QUICKMSG_MEMBERS))) {
			value = line + strlen(QUICKMSG_MEMBERS);
			
			members = calloc(1, ulen - strlen(QUICKMSG_MEMBERS) + 1);
			memcpy(members, value, ulen - strlen(QUICKMSG_MEMBERS));
		}
		if (!strncmp(line, QUICKMSG_TIME, strlen(QUICKMSG_TIME))) {
			value = line + strlen(QUICKMSG_TIME);
			havetime = true;
			acttime = atol(value);
		}
		if (!strncmp(line, QUICKMSG_GROUP, strlen(QUICKMSG_GROUP))) {
			value = line + strlen(QUICKMSG_GROUP);
			havegroupnr = true;
			groupnr = atoi(value);
		}
		if (ulen == 0) {
			char *p = line + len + 1;
			size_t pl = end - p;
			
			postdata = calloc(1, pl + 1);
			memcpy(postdata, p, pl);
		}
		
	}	
	if (post) {
		struct contact *c, *csend;
		struct post *p;
		
		if (!havetime)
			acttime = time(NULL);

		csend = quickmsg_db_contact_person_get_by_address(db, from);
		if (name) {
			contact_name_set(csend, name);
			quickmsg_db_contact_update(db, csend);
		}
		
		if (havegroupnr && owner) {
			c = quickmsg_db_contact_group_get_by_address_and_group(db, owner, groupnr);
			if (c) {
				size_t i;
				bool foundmember = false;
				
				for (i = 0; i < contact_members_get_size(c); i++) {
					char *member = contact_member_get(c, i);
					
					if (!strcmp(member, from))
						foundmember = true;
					
					free(member);
				}
				if (!foundmember)
					goto out;
			}
		} else {
			c = quickmsg_db_contact_person_get_by_address(db, from);
		}
		if (!c)
			goto out;
		
		if (havetime) {
			contact_time_lastact_set(c, acttime);
		}
		
		p = post_alloc();
		
		post_from_set(p, contact_id_get(csend));
		post_id_set(p, contact_id_get(c));
		post_time_set(p, acttime);
		post_text_set(p, postdata);
		
		if (a) {
			char *aname;
			
			aname = quickmsg_attachment_save(a);

			if (aname) {
				printf("Attachment: %s\n", aname);
				post_uri_set(p, aname);
			}
			free(aname);
		}
		
		quickmsg_db_post_insert(db, p);
		post_free(p);
		
		quickmsg_db_contact_update(db, c);
		contact_free(csend);
		contact_free(c);
	}
	if (group && havegroupnr && owner) {
		struct contact *c;
		bool update = true;
		
		c = quickmsg_db_contact_group_get_by_address_and_group(db, owner, groupnr);
		if (!c) {
			c = contact_alloc();
			contact_address_set(c, owner);
			contact_time_lastact_set(c, time(NULL));
			contact_group_set(c, groupnr);
			contact_type_set(c, CONTACT_TYPE_GROUP);
			update = false;
		}
	
		if (name) {
			contact_name_set(c, name);
		}
		if (members) {
			size_t i;
			
			contact_members_set(c, members);
			for (i = 0; i < contact_members_get_size(c); i++) {
				char *m = contact_member_get(c, i);
				struct contact *member;
				
				member = quickmsg_db_contact_person_get_by_address(db, m);
				if (!member) {
					member = contact_alloc();
					contact_address_set(member, m);
					contact_type_set(member, CONTACT_TYPE_PERSON);
					contact_keystat_set(member, CONTACT_KEYSTAT_VERIFIED);
					quickmsg_db_contact_insert(db, member);
				} else {
					contact_keystat_set(member, CONTACT_KEYSTAT_VERIFIED);
					quickmsg_db_contact_update(db, member);
				}
				
				free(m);
			}
		}
	
		if (update) {
			quickmsg_db_contact_update(db, c);
			printf("Update group\n");
		} else {
			quickmsg_db_contact_insert(db, c);
			printf("Add group\n");
		}
		contact_free(c);
	}
out:	
	free(name);
	free(owner);
	free(members);
	free(postdata);
}

void quickmsg_recv(struct quickmsg_db *db, struct gpg *gpg)
{
	struct mailbox *mailbox;
	uint32_t *uidlist;
	size_t uidlist_nr;
	int i;

	mailbox = mailbox_create(
	    settings_imap_host_get(),
	    0,
	    settings_imap_user_get(),
	    settings_imap_pass_get());

	if (!mailbox) {
		printf("Failed to open mailbox\n");
		return;
	}
	printf("Opened mailbox\n");

	mailbox_connect(mailbox);

	quickmsg_checkdir(mailbox);
	
	mailbox_chdir(mailbox, QUICKMSG_MAILBOX);
	
	mailbox_ls(mailbox, &uidlist, &uidlist_nr);

	for (i = 0; i < uidlist_nr; i++) {
		bool deleted;
		bool quickmsg;
		char *body;
		size_t length;
		char *from;
		int j;
		struct attachment_list *al;
		bool encrypted_found;
		struct attachment *encrypted = NULL;
		
		mailbox_get_id_is_deleted(mailbox, i, &deleted);
		if (deleted)
			continue;
			
		mailbox_get_id_is_quickmsg(mailbox, i, &quickmsg);
		if (!quickmsg)
			continue;
		
		mailbox_get_id_body(mailbox, i, &body, &length);
//		printf("Body: \n%s\n", body);

		message_parse_from(body, length, &from);
		printf("From: %s\n", from);

		message_parse_attachments(body, length, &al);
		for (j = 0; j < al->size; j++) {
//			printf("%d: %s\n", j, al->attachments[j]->type);
			
			if (!strcasecmp(al->attachments[j]->type, "pgp-keys")) {
				struct contact *c;
				gpg_add_key(gpg, al->attachments[j]);
				
				c = quickmsg_db_contact_person_get_by_address(
				    db, from);
				
				if (c) {
					printf("Contact exists\n");
				} else {
					c = contact_alloc();
					
					contact_address_set(c, from);
					contact_type_set(c, 
					    CONTACT_TYPE_PERSON);
					contact_keystat_set(c, 
					    CONTACT_KEYSTAT_RECEIVED);
					contact_time_lastact_set(c,
					    time(NULL));
					
					quickmsg_db_contact_insert(db, c);
					printf("Add contact\n");
				}

				contact_free(c);
			}
			
			if (!strcasecmp(al->attachments[j]->type, "pgp-encrypted")) {
				encrypted_found = true;
			}

			if (!strcasecmp(al->attachments[j]->type, "octet-stream")) {
				encrypted = al->attachments[j];
			}
			
		}
		if (encrypted_found && encrypted) {
			struct contact *c;
			printf("Received encrypted message\n");

			c = quickmsg_db_contact_person_get_by_address(db, from);
			if (c && (contact_keystat_get(c) == CONTACT_KEYSTAT_VERIFIED)) {
				struct attachment *plain =
				    gpg_decrypt_verify(gpg, encrypted, from);
				
				if (plain) {
					struct attachment_list *alp;
					struct attachment *qm = NULL;
					struct attachment *att = NULL;
				
					message_parse_attachments(
					    plain->data, plain->length, &alp);

					for (j = 0; j < alp->size; j++) {
						if (!strcasecmp(alp->attachments[j]->type, "quickmsg")) {
							qm = alp->attachments[j];
						} else if (!strcasecmp(alp->attachments[j]->type, "pgp-keys")) {
							gpg_add_key(gpg, alp->attachments[j]);
						} else {
							att = alp->attachments[j];
						}
					}
					if (qm)
						quickmsg_handle(db, from, qm, att);
					
					attachment_list_free(alp);
				} else {
					printf("Could not decrypt/verify message\n");
				}
			}
		}

		attachment_list_free(al);

		free(from);
		free(body);
		
		mailbox_delete_id(mailbox, i);
	}
	free(uidlist);

	mailbox_destroy(mailbox);

}

void quickmsg_key(struct quickmsg_db *db, struct gpg *gpg, char *toid)
{
	struct contact *c;
	char *me = settings_address_get();
	struct attachment *a;
	char *to;
	
	c = quickmsg_db_contact_get_by_id(db, atoi(toid));
	
	if (!c) {
		printf("Id %s not found\n", toid);
		return;
	}
	if (contact_type_get(c) != CONTACT_TYPE_PERSON) {
		printf("Not a person!\n");
		return;
	}
	
	to = contact_address_get(c);
	a = gpg_get_key(gpg, me);
	if (a) {
		struct attachment_list *l;
		
		l = attachment_list_alloc();
		if (!l)
			return;
			
		attachment_list_add(l, a);
		
		smtp_send(to, l, "multipart/mixed");
		
		attachment_list_free(l);
	}
}

void quickmsg_fingerprint(struct quickmsg_db *db, struct gpg *gpg, char *uid)
{
	struct contact *c;
	char *add;
	char *fp;
	
	c = quickmsg_db_contact_get_by_id(db, atoi(uid));
	
	if (!c) {
		printf("Id %s not found\n", uid);
		return;
	}
	add = contact_address_get(c);
	
	fp = gpg_get_fingerprint(gpg, add);
	if (fp) {
		printf("Fingerprint for contact %s (%s): %s\n",
		   uid, add, fp);
		free(fp);
	}
}

void quickmsg_verify(struct quickmsg_db *db, char *uid)
{
	struct contact *c;
	
	c = quickmsg_db_contact_get_by_id(db, atoi(uid));
	
	if (contact_type_get(c) == CONTACT_TYPE_PERSON) {
		if (contact_keystat_get(c) == CONTACT_KEYSTAT_RECEIVED) {
			contact_keystat_set(c, CONTACT_KEYSTAT_VERIFIED);
			quickmsg_db_contact_update(db, c);
			printf("Updated contact\n");
		}
	}
}

void quickmsg_remove(struct quickmsg_db *db, char *uid)
{
	struct contact *c;
	
	c = quickmsg_db_contact_get_by_id(db, atoi(uid));
	
	quickmsg_db_contact_delete(db, c);
	//todo remove key if person
	//todo remove groups if person
}

void quickmsg_add(struct quickmsg_db *db, char *add)
{
	struct contact *c;
	
	c = quickmsg_db_contact_person_get_by_address(db, add);
	if (c) {
		printf("Contact already exists\n");
		return;
	}
	
	c = contact_alloc();
	if (!c)
		return;

	contact_address_set(c, add);
	contact_type_set(c, CONTACT_TYPE_PERSON);
	contact_keystat_set(c, CONTACT_KEYSTAT_NONE);
	contact_time_lastact_set(c, time(NULL));
					
	quickmsg_db_contact_insert(db, c);
	printf("Add contact\n");
}

void quickmsg_read(struct quickmsg_db *db, char *uid)
{
	struct post_list *l;
	size_t i;
	
	l = quickmsg_db_post_get_by_id(db, atoi(uid));

	printf(post_print_header());
	for (i = 0; i < l->size; i++) {
		printf(post_print(l->posts[i]));
	}
	
	post_list_free(l);
}

void quickmsg_write(struct quickmsg_db *db, struct gpg *gpg, char *uid, char *text, char *filename)
{
	struct post *p;
	struct contact *c;
	char *msgtxt;
	bool isgroup;
	char groupnr[11];
	struct attachment *a, *af;
	struct attachment_list *al;
	struct attachment *plain;
	struct attachment *encrypted;
	void *plain_data;
	size_t plain_length;
	
	c = quickmsg_db_contact_get_by_id(db, atoi(uid));
	if (!c) {
		printf("Contact %s not found\n", uid);
		return;
	}
	
	isgroup = contact_type_get(c) == CONTACT_TYPE_GROUP;
	
	p = post_alloc();
	if (!p)
		goto free_c;
	
	post_id_set(p, contact_id_get(c));
	post_from_set(p, 1);
	post_text_set(p, text);
	if (filename)
		post_uri_set(p, filename);

	quickmsg_db_post_insert(db, p);

	post_free(p);

	sprintf(groupnr, "%d", contact_group_get(c));

	asprintf(&msgtxt,
	    QUICKMSG_ACTION QUICKMSG_ACTION_POST "\n"
	    QUICKMSG_NAME "%s\n"
	    QUICKMSG_TIME "%ld\n"
	    "%s%s%s"
	    "%s%s%s"
	    "\n%s",
	    settings_name_get(),
	    time(NULL),
	    isgroup ? QUICKMSG_OWNER : "",
	    isgroup ? contact_address_get(c) : "",
	    isgroup ? "\n" : "",
	    isgroup ? QUICKMSG_GROUP : "",
	    isgroup ? groupnr : "",
	    isgroup ? "\n" : "",
	    text);

	a = attachment_alloc();
	if (!a)
		goto free_c;
	
	attachment_data_set(a, msgtxt, strlen(msgtxt));
	free(msgtxt);
	attachment_type_set(a, "quickmsg");
	attachment_basetype_set(a, "application");
	
	al = attachment_list_alloc();
	if (!al)
		goto free_a;
	attachment_list_add(al, a);

	if (filename) {
		af = attachment_file_create(filename);
		if (af)
			attachment_list_add(al, af);
	}
	
	message_multipart_create(NULL, al, &plain_data, &plain_length, "multipart/mixed");
	plain = attachment_alloc();
	attachment_data_set(plain, plain_data, plain_length);
	attachment_type_set(plain, "mixed");
	attachment_basetype_set(plain, "multipart");
	free(plain_data);
	
	attachment_list_remove(al, 0);
	attachment_list_free(al);	

	if (isgroup) {
		size_t i;
		
		for (i = 0; i < contact_members_get_size(c); i++) {
			char *member = contact_member_get(c, i);
			if (!strcmp(member, settings_address_get()))
				continue;
			
			encrypted = gpg_encrypt_sign(gpg, plain, member);
			if (encrypted) {
				struct attachment_list *ml = attachment_list_alloc();
			
				attachment_list_add(ml, pgp_mime_id_get());
				attachment_list_add(ml, encrypted);

				smtp_send(member, ml, "multipart/encrypted");
				attachment_list_free(ml);
			}
		}
	} else {
		encrypted = gpg_encrypt_sign(gpg, plain, contact_address_get(c));
		if (encrypted) {
			struct attachment_list *ml = attachment_list_alloc();
			
			attachment_list_add(ml, pgp_mime_id_get());
			attachment_list_add(ml, encrypted);
			smtp_send(contact_address_get(c), ml, "multipart/encrypted");
			attachment_list_free(ml);
		}
	}	

	attachment_free(plain);
	
free_a:
	attachment_free(a);
free_c:
	contact_free(c);
}

void quickmsg_group(struct quickmsg_db *db, struct gpg *gpg, struct contact *g)
{
	char *msgtxt;
	char groupnr[11];
	struct attachment *a;
	struct attachment_list *al;
	struct attachment *plain;
	struct attachment *encrypted;
	void *plain_data;
	size_t plain_length;
	size_t i;
	
	if (!contact_type_get(g) == CONTACT_TYPE_GROUP)
		return;
	if (strcmp(contact_address_get(g), settings_address_get()))
		return;
	
	sprintf(groupnr, "%d", contact_group_get(g));

	asprintf(&msgtxt,
	    QUICKMSG_ACTION QUICKMSG_ACTION_GROUP "\n"
	    QUICKMSG_NAME "%s\n"
	    QUICKMSG_TIME "%ld\n"
	    QUICKMSG_OWNER "%s\n"
	    QUICKMSG_GROUP "%s\n"
	    QUICKMSG_MEMBERS "%s\n"
	    "\n",
	    contact_name_get(g),
	    time(NULL),
	    contact_address_get(g),
	    groupnr,
	    contact_members_get(g));

	a = attachment_alloc();
	if (!a)
		goto err;
	
	attachment_data_set(a, msgtxt, strlen(msgtxt));
	free(msgtxt);
	attachment_type_set(a, "quickmsg");
	attachment_basetype_set(a, "application");
	
	al = attachment_list_alloc();
	if (!al)
		goto free_a;
	attachment_list_add(al, a);
	
	for (i = 0; i < contact_members_get_size(g); i++) {
		char *member = contact_member_get(g, i);
		
		attachment_list_add(al, gpg_get_key(gpg, member));
		
		free(member);
	}

	message_multipart_create(NULL, al, &plain_data, &plain_length, "multipart/mixed");
	plain = attachment_alloc();
	attachment_data_set(plain, plain_data, plain_length);
	attachment_type_set(plain, "mixed");
	attachment_basetype_set(plain, "multipart");
	
	attachment_list_remove(al, 0);
	attachment_list_free(al);	

	for (i = 0; i < contact_members_get_size(g); i++) {
		char *member = contact_member_get(g, i);
		if (!strcmp(member, settings_address_get()))
			continue;
		
		encrypted = gpg_encrypt_sign(gpg, plain, member);
		if (encrypted) {
			struct attachment_list *ml = attachment_list_alloc();
		
			attachment_list_add(ml,pgp_mime_id_get());
			attachment_list_add(ml, encrypted);

			smtp_send(member, ml, "multipart/encrypted");
			attachment_list_free(ml);
		}
	}

	attachment_free(plain);
	
free_a:
	attachment_free(a);
err:
	return;
}

void quickmsg_add_group(struct quickmsg_db *db, char *name)
{
	struct contact *g;
	int nr;
	char *add = settings_address_get();
	
	srandom(time(NULL));
	do {
		nr = random() & 0xffff;
		g = quickmsg_db_contact_group_get_by_address_and_group(db, add, nr);
		if (!g)
			break;
		contact_free(g);
	} while (1);
	
	g = contact_alloc();
	if (!g)
		return;

	contact_address_set(g, add);
	contact_name_set(g, name);
	contact_type_set(g, CONTACT_TYPE_GROUP);
	contact_keystat_set(g, CONTACT_KEYSTAT_NONE);
	contact_time_lastact_set(g, time(NULL));
	contact_group_set(g, nr);
	contact_members_set(g, add);
					
	quickmsg_db_contact_insert(db, g);
	contact_free(g);
	printf("Add group %d\n", nr);
}

void quickmsg_add_group_member(struct quickmsg_db *db, struct gpg *gpg, char *id, char *name)
{
	struct contact *g;
	struct contact *c;
	char *add = settings_address_get();
	size_t i;
	bool exists = false;
	
	g = quickmsg_db_contact_get_by_id(db, atoi(id));
	if (!g) {
		printf("Group not found: %d\n", atoi(id));
		return;
	}
	if (contact_type_get(g) != CONTACT_TYPE_GROUP) {
		printf("Not a group\n");
		return;
	}
	if (strcmp(contact_address_get(g), add)) {
		printf("Not my group\n");
		return;
	}
	
	c = quickmsg_db_contact_person_get_by_address(db, name);
	if (!c) {
		printf("Contact not found\n");
		return;
	}
	
	if (contact_keystat_get(c) != CONTACT_KEYSTAT_VERIFIED) {
		printf("Contact not verified\n");
		return;
	}
	
	for (i = 0; i < contact_members_get_size(g); i++) {
		char *m;
		
		m = contact_member_get(g, i);
		
		printf("%s %s\n", name, m);
		if (!strcmp(name, m)) {
			exists = true;
		}
		
		free(m);
	}
	
	if (!exists)
		contact_members_add(g, name);
	
	quickmsg_db_contact_update(db, g);
	quickmsg_group(db, gpg, g);
	
	printf("Update group members: %s\n", contact_members_get(g));

	contact_free(g);
}

void quickmsg_remove_group_member(struct quickmsg_db *db, struct gpg *gpg, char *id, char *name)
{
	struct contact *g;
	struct contact *c;
	char *add = settings_address_get();
	bool exists = false;
	
	g = quickmsg_db_contact_get_by_id(db, atoi(id));
	if (!g) {
		printf("Group not found: %d\n", atoi(id));
		return;
	}
	if (contact_type_get(g) != CONTACT_TYPE_GROUP) {
		printf("Not a group\n");
		return;
	}
	if (strcmp(contact_address_get(g), add)) {
		printf("Not my group\n");
		return;
	}
	if (!strcmp(name, add)) {
		printf("Can not remove myself\n");
		return;
	}
	
	c = quickmsg_db_contact_person_get_by_address(db, name);
	if (!c) {
		printf("Contact not found\n");
		return;
	}
	
	contact_members_remove(g, name);
	
	if (!exists)
		contact_members_add(g, name);
	
	quickmsg_db_contact_update(db, g);
	quickmsg_group(db, gpg, g);
	
	printf("Update group members: %s\n", contact_members_get(g));

	contact_free(g);
}

int main(int argc, char **argv)
{
	struct gpg *gpg;
	struct quickmsg_db *db;

	db = quickmsg_db_alloc();
	quickmsg_check_me(db);

	gpg = gpg_alloc();

	if (argc >= 2) {
		if (!strcmp(argv[1], "list")) {
			quickmsg_list(db);
			goto done;
		}
		
		if (!strcmp(argv[1], "add") && argc >= 3) {
			quickmsg_add(db, argv[2]);
			goto done;
		}

		if (!strcmp(argv[1], "verify") && argc >= 3) {
			quickmsg_verify(db, argv[2]);
			goto done;
		}

		if (!strcmp(argv[1], "remove") && argc >= 3) {
			quickmsg_remove(db, argv[2]);
			goto done;
		}
		
		if (!strcmp(argv[1], "fingerprint") && argc >= 3) {
			quickmsg_fingerprint(db, gpg, argv[2]);
			goto done;
		}

		if (!strcmp(argv[1], "key") && argc >= 3) {
			quickmsg_key(db, gpg, argv[2]);
			goto done;
		}

		if (!strcmp(argv[1], "read") && argc >= 3) {
			quickmsg_read(db, argv[2]);
			goto done;
		}
		
		if (!strcmp(argv[1], "write") && argc >= 4) {
			if (argc == 4)
				quickmsg_write(db, gpg, argv[2], argv[3], NULL);
			else
				quickmsg_write(db, gpg, argv[2], argv[3], argv[4]);
			goto done;
		}
		
		if (!strcmp(argv[1], "add_group") && argc >= 3) {
			quickmsg_add_group(db, argv[2]);
			goto done;
		}
		if (!strcmp(argv[1], "add_member") && argc >= 4) {
			quickmsg_add_group_member(db, gpg, argv[2], argv[3]);
			goto done;
		}
		if (!strcmp(argv[1], "remove_member") && argc >= 4) {
			quickmsg_remove_group_member(db, gpg, argv[2], argv[3]);
			goto done;
		}
	}

	quickmsg_recv(db, gpg);

done:
	gpg_free(gpg);
	quickmsg_db_free(db);
	
	return 0;
}
